"""Plotting functions for Flow Matching."""

from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np
from jaxtyping import Array, Float


def plot_data(
    z0: Float[Array, "nsamples dim"], z1: Float[Array, "nsamples dim"]
) -> None:
    """Plot initial data points."""
    plt.figure(figsize=(8, 5))
    plt.scatter(z0[:, 0], z0[:, 1], alpha=0.3, label="z0", s=10)
    plt.scatter(z1[:, 0], z1[:, 1], alpha=0.3, label="z1", s=10)
    plt.legend()
    plt.axis("equal")
    plt.savefig("fig/data.pdf")


def plot_init_traj(
    z0: Float[Array, "nsamples dim"],
    z1: Float[Array, "nsamples dim"],
    zt: Float[Array, "nsamples dim"],
    n: int = 1000,
) -> None:
    """Plot initial trajectories.

    Args:
        z0: Source points.
        z1: Target points.
        zt: Trajectory points.
        n: Number of trajectories to plot.
    """
    plt.figure(figsize=(8, 5))
    assert len(z0) == len(z1) == len(zt)
    idx = np.random.choice(len(z0), n)
    z0 = z0[idx]  # + np.array([-5., 0.])
    z1 = z1[idx]  # + np.array([5., 0.])
    plt.scatter(z0[:, 0], z0[:, 1], alpha=0.4, label="z0", s=10)
    plt.scatter(z1[:, 0], z1[:, 1], alpha=0.4, label="z1", s=10)
    # plt.plot(zt[:, 0], zt[:, 1], 'go', label='zt')
    plt.legend()
    plt.savefig("fig/init_traj.pdf")


def plot_loss_curve(loss_curve: List[float], **kwargs) -> None:
    """Plot the loss curve of the neural vector field.

    Args:
        loss_curve: Loss at each iteration.
        **kwargs: Figure options.
    """
    plt.figure(figsize=(8, 5))
    plt.plot(loss_curve)
    plt.xlabel("iterations")
    plt.ylabel("loss")
    plt.savefig(f"fig/{kwargs.get('title', 'loss_curve')}.pdf")


def plot_traj(
    traj: Float[Array, "batch_size ntraj dim"],
    x1: Float[Array, "batch_size dim"],
    ntraj: Optional[int] = None,
    **fig_kwargs,
) -> None:
    """Plot a given number of trajectories.

    Args:
        traj: Array of trajectories (batch_size, ntraj, dim).
        x1: Target points.
        ntraj: Number of trajectories to plot.
        **fig_kwargs: Figure options.
    """
    if ntraj is None:
        ntraj = traj.shape[1]
    labels = fig_kwargs.get("labels", None)
    plt.figure(figsize=(6, 6))
    plt.tight_layout()
    for i in range(ntraj):
        if labels is not None:
            idx = np.where(labels == 0)[0]
            plt.plot(
                traj[:, idx[i], 0],
                traj[:, idx[i], 1],
                alpha=0.2,
                color="blue",
                zorder=1,
            )
            idx = np.where(labels == 1)[0]
            plt.plot(
                traj[:, idx[i], 0], traj[:, idx[i], 1], alpha=0.2, color="red", zorder=1
            )
        else:
            plt.plot(traj[:, i, 0], traj[:, i, 1], alpha=0.2, color="olive", zorder=1)
    plt.scatter(
        traj[0, :, 0],
        traj[0, :, 1],
        label=r"$z_0$",
        s=4,
        alpha=0.5,
        zorder=2,
        color="black",
    )
    plt.scatter(
        traj[-1, :, 0],
        traj[-1, :, 1],
        label=r"$T(z_0)$",
        s=4,
        alpha=0.2,
        zorder=3,
        color="blue",
    )
    plot_target = fig_kwargs.get("plot_target", False)
    if plot_target:
        plt.scatter(
            x1[:, 0],
            x1[:, 1],
            label=r"$z_1$",
            s=4,
            alpha=0.2,
            zorder=2,
            color="red",
        )
    plt.legend(fontsize=12)
    plt.axis("equal")
    plt.savefig(f"fig/{fig_kwargs.get('title', 'traj')}.pdf")



from typing import Tuple
import numpy as np


def bimodal_gaussian(
    num_samples: int,
    mean_1: np.ndarray = np.array([0.0, 0.0]),
    mean_2: np.ndarray = np.array([3.0, 0.0]),
    sigma_1: float = 0.2,
    sigma_2: float = 0.1,
) -> Tuple[np.ndarray, np.ndarray]:
    """Generate bimodal gaussian data

    Args:
        num_samples: Number of samples to generate
        mean_1: Mean of the first gaussian
        mean_2: Mean of the second gaussian
        sigma_1: Std of the first gaussian
        sigma_2: Std of the second gaussian

    Returns:
        Data and labels
    """
    x = np.concatenate(
        [
            np.random.randn(num_samples // 2, 2) * sigma_1 + mean_1,
            np.random.randn(num_samples // 2, 2) * sigma_2 + mean_2,
        ],
        axis=0,
    )
    labels = np.concatenate([np.zeros(num_samples // 2), np.ones(num_samples // 2)])
    perm = np.random.permutation(len(x))
    return x[perm], labels[perm]



from jaxfm.tasks.bimodal_gaussians import bimodal_gaussian
from jaxfm.tasks.two_moons import two_moons
from jaxfm.tasks.eight_gaussians import eight_gaussian

def get_data(num_samples: int, task_name: str, **kwargs):
    """Return data for the given task

    Args:
        num_samples: Number of samples to generate
        task_name: Name of the task
        **kwargs: Task specific arguments

    Raises:
        ValueError: If the task is unknown

    Returns:
        Data and labels
    """
    if task_name == "bimodal_gaussian":
        return bimodal_gaussian(num_samples, **kwargs)
    if task_name == "two_moons":
        return two_moons(num_samples, **kwargs)
    if task_name == "eight_gaussian":
        return eight_gaussian(num_samples, **kwargs)
    else:
        raise ValueError(f"Unknown task {task_name}")



import equinox as eqx
from jaxtyping import Array, Float
import jax.numpy as jnp
import jax

class MLP(eqx.Module):
    """Module to learn the score function using a MLP
    The output of the model is the a vector of the same size as the input theta
    """

    layers: list

    def __init__(
        self,
        key: Array,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        n_layers: int = 3,
    ):
        keys = jax.random.split(key, n_layers)
        self.layers = [
            eqx.nn.Linear(input_dim, hidden_dim, key=keys[0]),
            jax.nn.relu,
        ]
        for i in range(n_layers - 2):
            self.layers.extend(
                [
                    eqx.nn.Linear(hidden_dim, hidden_dim, key=keys[i + 1]),
                    jax.nn.tanh,
                ]
            )
        self.layers.extend(
            [
                eqx.nn.Linear(hidden_dim, output_dim, key=keys[-1]),
            ]
        )

    def __call__(
        self, x_t: Float[Array, "dim_x_t"], t: Float[Array, "dim_t"]
    ) -> Float[Array, "dim_"]:
        inputs = jnp.concatenate([x_t, t], axis=-1)
        for layer in self.layers:
            inputs = layer(inputs)
        return inputs





"""Base class for flow models."""

from abc import abstractmethod
import jax

import jax.numpy as jnp
from typing import Tuple

from jaxtyping import Array, Float


class Flow:
    """Base class for flow models."""

    def __init__(self, model=None, num_steps=100, law: str= "uniform") -> None:
        self.model = model
        self.N = num_steps
        self.law = law

    @abstractmethod
    def get_train_tuple(
        self,
        z0: Float[Array, "batch_size dim"],
        z1: Float[Array, "batch_size dim"],
        key: int,
        **kwargs,
    ) -> Tuple[
        Float[Array, "batch_size dim"],
        Float[Array, "batch_size 1"],
        Float[Array, "batch_size dim"],
    ]:
        """Get the interpolated samples and the target."""
        pass

    def sample_ode(self, z0: jnp.ndarray, N=None) -> jnp.ndarray:
        """Sample the ODE"""
        if N is None:
            N = self.N

        dt = 1 / N
        traj = []
        z = z0
        batch_size = z0.shape[0]

        @jax.jit
        def euler_step(z, i, dt):
            t = jnp.ones((batch_size, 1)) * i * dt
            pred = jax.vmap(self.model)(z, t)
            return z + pred * dt

        traj.append(z)
        for i in range(N):
            z = euler_step(z, i, dt)
            traj.append(z)
        return jnp.asarray(traj)



from jaxfm.flows.base import Flow
from jaxfm.utils.time_scheduler import get_t_samples
from jaxtyping import Array, Float
from typing import Tuple
import jax


class CFM(Flow):
    """Conditional Flow model
    Implements Conditional Flow Matching for the Optimal Transport objective i.e. constant vector field.

    Reference:
    Flow Matching for Generative Modeling (https://arxiv.org/abs/2210.02747)
    """

    def __init__(self, model=None, num_steps=100, law: str = "uniform", noise=0.01) -> None:
        """Initialize the Flow Matching procedure with the model and the number of steps

        Args:
            model : The model to learn the vector field
            num_steps : The number of steps for the Euler integration
            noise : The noise to add to the samples
        """
        self.model = model  # The model to learn the vector field
        self.N = num_steps  # The number of steps for the Euler integration
        self.sigma = noise  # The noise to add to the samples
        self.law = law # The law for the time sampling

    def get_train_tuple( # type: ignore
        self,
        z0: Float[Array, "batch_size dim"],
        z1: Float[Array, "batch_size dim"],
        key: int,
    ) -> Tuple[
        Float[Array, "batch_size dim"],
        Float[Array, "batch_size 1"],
        Float[Array, "batch_size dim"],
    ]:
        """Interpolate the samples and get the target.

        Args:
            z0: Source samples
            z1: Target samples
            key: Random key

        Returns:
            Interpolated samples, time samples, target
        """
        t = get_t_samples(key, z0.shape[0], law=self.law)
        z_t = (
            z0 * (1 - t)
            + z1 * t
            + self.sigma * jax.random.normal(key, (z0.shape[0], z0.shape[1]))
        )
        return z_t, t, z1 - z0
